import argparse
import os
import sys

from enum import StrEnum
from importlib import metadata
from syllogistic.watcher import BuildWatcher


# test.lua.tic -> test, .lua.tic
def splitext2(path):
  filename, ext = os.path.splitext(path)

  if len(filename) > 1:
    filename2 = os.path.splitext(filename)

    if len(filename2) > 1:
      ext = filename2[1] + ext
      filename = filename2[0]

  return (filename, ext)

def main():
  version = metadata.version('syllogistic')

  argparser = argparse.ArgumentParser(
    prog='syllogistic',
    description=f'Build system for TIC-80 fantasy computer carts. v{version}'
  )
  argparser.add_argument('input', nargs='+', help='input file')
  argparser.add_argument('-o', '--out', default=os.getcwd(), metavar='str', help='output file or directory')
  argparser.add_argument('-s', '--strict', action='store_const', const=True, help='enable strict checking')
  argparser.add_argument('-w', '--watch', action='store_const', const=True, help='watch for file changes')
  argparser.add_argument('-vv', '--verbose', action='store_const', const=True, help='verbose output (where relevant)')

  args = argparser.parse_args()
  file = args.input[0]
  filename = splitext2(args.input[0])
  outfile = os.path.join(args.out, filename[0] + '.build' + filename[1]) if os.path.isdir(args.out) else args.out

  watcher = BuildWatcher(version, file, outfile, args.strict, args.verbose)
 
  if args.watch:
    watcher.start()
  else:
    watcher.build()
    print('')

if __name__ == '__main__':
  main()



import os
import pytest
import unittest

from syllogistic.processor import DEFAULT_TIC_80_PATH, ProcessorContext, ProcessorException, DynamicAssignmentProcessorException, ModuleNotFoundProcessorException, InvalidSyntaxProcessorException, InvalidStringEscapeProcessorException, process_lines


class TestRequire(unittest.TestCase):
    def test_require(self):
      code = "require('tests/input/sample')"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
      result = process_lines(context)

      assert result == """function __syllogistic_require_1() function sample()
  trace('sample')
end end
__syllogistic_require_1()"""

    def test_require_escape(self):
      code = "require('tests/input/s\\'ample')"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
      result = process_lines(context)

      assert result == """function __syllogistic_require_1() function sample()
  trace('sample')
end end
__syllogistic_require_1()"""

    def test_require_init(self):
      code = "require('tests/input')"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
      result = process_lines(context)

      assert result == """function __syllogistic_require_1() function init() end end
__syllogistic_require_1()"""

    def test_require_assign(self):
      code = "variable = require('tests/input/sample')"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
      result = process_lines(context)

      assert result == """function __syllogistic_require_1() function sample()
  trace('sample')
end end
variable = __syllogistic_require_1()"""

    def test_require_assign_semicolon(self):
      code = "variable = require('tests/input/sample'); test = 3"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
      result = process_lines(context)

      assert result == """function __syllogistic_require_1() function sample()
  trace('sample')
end end
variable = __syllogistic_require_1(); test = 3"""

    def test_require_package_path(self):
      code = """variable = 3
package.path=package.path ..';tests/input/?.lua'
require('sample')"""
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
      result = process_lines(context)

    def test_require_package_path_depth(self):
      code = """package.path=package.path ..';tests/input/?.lua'
require('require')"""
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
      result = process_lines(context)

      assert result == """function __syllogistic_require_2() function sample()
  trace('sample')
end end
function __syllogistic_require_1() __syllogistic_require_2() end
__syllogistic_require_1()"""
      expected = list(DEFAULT_TIC_80_PATH)
      expected.append('tests/input/?.lua')
      assert context.requires.path == expected

    def test_require_package_path_depth_already_loaded(self):
      ModuleNotFoundProcessorException
      code = """package.path=package.path ..';tests/input/?.lua'
require('sample')
require('require-no-path')"""
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
      result = process_lines(context)

      assert result == """function __syllogistic_require_1() function sample()
  trace('sample')
end end
function __syllogistic_require_2() __syllogistic_require_1() end
__syllogistic_require_1()
__syllogistic_require_2()"""
      expected = list(DEFAULT_TIC_80_PATH)
      expected.append('tests/input/?.lua')
      assert context.requires.path == expected

    def test_require_package_path_after(self):
      code = """variable = 3
package.path='tests/input/?.lua;'.. package.path
require('sample')"""
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
      result = process_lines(context)

      assert result == """function __syllogistic_require_1() function sample()
  trace('sample')
end end
variable = 3
__syllogistic_require_1()"""
      expected = ['tests/input/?.lua']
      expected.extend(DEFAULT_TIC_80_PATH)
      assert context.requires.path == expected

    def test_require_package_path_in_line(self):
      code = """variable = 3
package.path='tests/input/?.lua;'.. package.path; variable=7
require('sample')"""
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
      result = process_lines(context)

      assert result == """function __syllogistic_require_1() function sample()
  trace('sample')
end end
variable = 3
variable=7
__syllogistic_require_1()"""
      expected = ['tests/input/?.lua']
      expected.extend(DEFAULT_TIC_80_PATH)
      assert context.requires.path == expected

    def test_require_package_path_in_line_space(self):
      code = """variable = 3
package.path='tests/input/?.lua;'.. package.path variable=7
require('sample')"""
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
      result = process_lines(context)

      assert result == """function __syllogistic_require_1() function sample()
  trace('sample')
end end
variable = 3
variable=7
__syllogistic_require_1()"""
      expected = ['tests/input/?.lua']
      expected.extend(DEFAULT_TIC_80_PATH)
      assert context.requires.path == expected

    def test_require_set_absolute(self):
      path = os.path.abspath('tests/input/sample.lua')

      # Windows-style paths :-/
      if path.find('\\'):
        path = path.replace('\\', '\\\\')

      code = f"""package.path='?;?.lua'
require('{path}')"""
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
      result = process_lines(context)

      assert result == """function __syllogistic_require_1() function sample()
  trace('sample')
end end
__syllogistic_require_1()"""
      assert context.requires.path == ['?', '?.lua']

    def test_require_no_quotes(self):
      with pytest.raises(ProcessorException):
        code = "require(tests/input/sample)"
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
        result = process_lines(context)

    def test_require_no_empty(self):
      with pytest.raises(ModuleNotFoundProcessorException):
        code = "require()"
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
        result = process_lines(context)

    # Technically this should be allowed but it's easier to deny it than supported.
    def test_require_no_dynamic_require(self):
      with pytest.raises(ProcessorException):
        code = "require('x'..'y')"
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
        result = process_lines(context)

    def test_require_invalid_string_escape(self):
      with pytest.raises(InvalidStringEscapeProcessorException):
        code = """require('\\x1234sample')"""
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
        result = process_lines(context)

    def test_require_package_path_in_line_missing_semicolon_start(self):
      with pytest.raises(ModuleNotFoundProcessorException):
        code = """package.path=package.path.. 'tests/input/?.lua'
  require('sample')"""
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
        result = process_lines(context)

    def test_require_package_path_missing_semicolon_end(self):
      with pytest.raises(ModuleNotFoundProcessorException):
        code = """package.path='tests/input/?.lua'.. package.path
  require('sample')"""
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
        result = process_lines(context)

    def test_require_package_path_unterminated(self):
      with pytest.raises(InvalidSyntaxProcessorException):
        code = """package.path='tests/input/?.lua".. package.path
  require('sample')"""
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
        result = process_lines(context)

    def test_require_package_path_runon(self):
      with pytest.raises(InvalidSyntaxProcessorException):
        code = """package.path='tests/input/?.lua".. package.path = 7
  require('sample')"""
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
        result = process_lines(context)

    def test_require_package_path_unallowed_variable(self):
      with pytest.raises(DynamicAssignmentProcessorException):
        code = """package.path=packg..'tests/input/?.lua"
  require('sample')"""
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
        result = process_lines(context)

    def test_require_package_path_invalid_string_escape(self):
      with pytest.raises(InvalidStringEscapeProcessorException):
        code = """package.path='tests\\x1234/input/?.lua'
  require('sample')"""
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
        result = process_lines(context)

    def test_require_package_path_depth_not_found(self):
      with pytest.raises(ModuleNotFoundProcessorException):
        code = """package.path=package.path ..';tests/input/?.lua'
  require('require-no-path')"""
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True, verbose=True);
        result = process_lines(context)



import os
import re

from enum import StrEnum
from pathlib import Path


CWD = os.getcwd()

# https://is.gd/4s1GdR
# tl;dr: I chose it, I published it, it's baked in, it's too late to change it! Thank you for understanding.
PACKAGES_DIR = '.tic_pkgs'
PACKAGE_PATH_VAR = 'package.path'
DEFAULT_TIC_80_PATH = [
  CWD + '/lua/?.lua',
  CWD + '/lua/?/init.lua',
  CWD + '/?.lua',
  CWD + '/?/init.lua',
  CWD + '/../share/lua/5.3/?.lua',
  CWD + '/../share/lua/5.3/?/init.lua',
  # Not understanding these values... they're functionally equivalent to the
  # upper ones, since TIC-80 doesn't seem to interpret them as relative to the
  # loaded cart's location.
  #'./?.lua',
  #'./?/init.lua'
]

class MagicComment(StrEnum):
  """
    Supported 'magic' comment types used by this build system
  """
  GROUP_START='#group:start'
  GROUP_END='#group:end'
  IMPORT='#import'
  INCLUDE='#include'
  EXPORTS='#exports'

class ProcessorGroup:
  def __init__(self, name, disabled):
    self.name = name
    self.disabled = disabled

class ProcessorRequires:
  def __init__(self, context, root=None):
    self.context = context
    self.root = root if root else self
    self.semicolon = False

    if self.root == self:
      self.index = 0
      self.loaded = dict()
      self.output = []

    if 'LUA_PATH' in os.environ:
      self.path = os.environ['LUA_PATH'].trim().split(';')
    else:
      # This is what TIC-80 spits out as its default package.path value
      self.path = list(DEFAULT_TIC_80_PATH)

  def write(self):
    return '\n'.join(self.root.output)

  def exists(self):
    return self.root == self and len(self.loaded) > 0

  def get_name(self, index):
    return f'__syllogistic_require_' + str(index)

  def add_require(self, module):
    # As per the docs, if we've already loaded it we're to return it, even if we
    # wouldn't be able to find it in the current file's package.path
    # https://www.lua.org/pil/8.1.html
    if not module in self.root.loaded:
      found_path = None
      paths_tried = []

      for path in self.path:
        path = path.replace('?', module)
        paths_tried.append(f'no file `{path}`')

        if os.path.exists(path):
          found_path = path
          break

      if not found_path:
        trace = '\n' + '\n'.join(paths_tried) if self.context.verbose else ''
        raise ModuleNotFoundProcessorException(f'Module {module} not found{trace}', self.context)

      self.root.index += 1
      self.root.loaded[module] = self.root.index

      # Technically a race, since the file could be deleted after we tested
      # existence but not something to really care about for this program...
      with open(found_path, 'r') as file:
        lines = file.read().split('\n')
        index = self.root.index
        content = process_lines(ProcessorContext(
          found_path,
          lines,
          context=self.context
        ))
        self.root.output.append(f'function {self.get_name(index)}() {content} end')

    return self.get_name(self.root.loaded[module])

  def get_path(self):
    return ';'.join(self.path)

  def set_path(self, path):
    stripped = path.strip()
    self.semicolon = stripped.endswith(';')
    self.path = [part for part in stripped.split(';') if part != '']

  def append_path(self, path):
    stripped = path.strip()
    semistart = stripped.startswith(';')
    semicolon = stripped.endswith(';')
    parts = stripped.split(';')

    # Look, if you ended your last bit without a semicolon that's not my fault
    if not self.semicolon and not semistart:
      self.path[-1] = self.path[-1] + parts[0]
      parts = parts[1:]

    self.semicolon = semicolon
    self.path.extend([part for part in parts if part != ''])

class ProcessorContext:
  requires = None
  exports = set()
  groups = []
  disabled = False
  line = None
  number = -1

  # Refactor to figure out how to do this in a way that's not a bunch of optional arguments?
  def __init__(
    self,
    path,
    lines,
    strict=False,
    files=None,
    aka=None,
    included=None,
    excluded=None,
    watched=False,
    root=None,
    verbose=False,
#    requires=None,
    context=None
  ):
    self.path = path
    self.lines = lines
    self.aka = aka

    if context:
      self.strict = context.strict
      self.files = context.files
      self.included = set(context.included) if included is None else included
      self.excluded = set(context.included) if excluded is None else excluded
      self.watched = context.watched
      self.is_root = False
      self.root = context.root
      self.verbose = context.verbose
      self.requires = ProcessorRequires(self, context.requires.root)
    else:
      self.strict = strict
      self.files = set() if files is None else files
      self.aka = aka
      self.included = set() if included is None else included
      self.excluded = set() if excluded is None else excluded
      self.watched = watched
      self.is_root = root is None
      self.root = os.path.dirname(os.path.abspath(path)) if root is None else root
      self.verbose = verbose
      self.requires = ProcessorRequires(self)

    # Remove last newline if present
    if self.lines[-1] == '':
      self.lines.pop()

  def current_group(self):
    return self.groups[-1]

  def _update_status(self):
    # We might be checking his often so it makes sense to do it on group
    # add/remove rather than on every access to the enabled prop
    disabled = False

    for group in self.groups:
      if group.disabled:
        disabled = True
        break

    self.disabled = disabled

  def _is_disabled(self, name, default):
    if name in self.included:
      return False
    elif name in self.excluded:
      return True

    return default

  def push_group(self, name):
    if name.startswith('-'):
      name = name[1:]
      self.groups.append(ProcessorGroup(name, self._is_disabled(name, True)))
    else:
      if name.startswith('+'):
        name = name[1:]
      self.groups.append(ProcessorGroup(name, self._is_disabled(name, False)))

    self._update_status()

  def pop_group(self, name):
    if name != self.current_group().name:
      raise InvalidSyntaxProcessorException(f"Invalid group '{name}' nesting", self)

    self.groups.pop()
    self._update_status()

class ProcessorException(Exception):
  def __init__(self, message, context):
    super().__init__(message)

    self.file = context.path
    self.line = context.number

class ModuleNotFoundProcessorException(ProcessorException):
  pass

class DynamicModuleProcessorException(ProcessorException):
  pass

class DynamicAssignmentProcessorException(ProcessorException):
  pass

class InvalidSyntaxProcessorException(ProcessorException):
  pass

class InvalidStringEscapeProcessorException(ProcessorException):
  pass



def raise_concern(message, context):
  if context.strict:
    raise ProcessorException(message, context)

  prefix = '\n' if context.watched else ''
  print(prefix + 'Warning: ' + message)

# Important stuff
def resolve_package(context, path, mangled):
  name = path.split('/')[0]
  current = Path(context.root)
  drive = os.path.splitdrive(path)

  # Allowing C:// style abs paths does seem to accidentally work, but I don't
  # want to be on the hook for continuing to support anything other than *nix-
  # style pathing that, if needed, is easy to convert to the local system's.
  if drive[0] != '':
    raise ProcessorException(f"Invalid package name '{name}'", context)

  # Check for reserved names
  if re.match(r'^\W', path):
    raise ProcessorException(f"Invalid package name '{name}'", context)

  # There's probably some better way of walking up and doing this?
  while not current.joinpath(PACKAGES_DIR, path).exists() and current.name:
    current = current.parent

  # Fall back to the user's home directory
  if not current.joinpath(PACKAGES_DIR, path).exists():
    current = Path.home()

  if not current.joinpath(PACKAGES_DIR, path).exists():
    # Unmangle our sugar
    if mangled:
      name = re.sub(r'[/\\]index.lua', '', name)
    raise ModuleNotFoundProcessorException(f"Unable to resolve package '{name}'", context)

  # os.path.abspath is to match our usages of that elsewhere so things are consistent
  return os.path.abspath(str(current.joinpath(PACKAGES_DIR, path).absolute()))

def process_include(context, args):
  if not context.disabled:
    path = ''
    terminated = 1
    included = context.included
    excluded = context.excluded
    aka = None
    mangled = False

    if args[0].startswith("'") and not args[0].endswith('"'):
      terminated = -1

      for i in range(len(args)):
        path += args[i]

        if args[i].endswith("'"):
          terminated = i + 1
          break

      if terminated == -1:
        raise InvalidSyntaxProcessorException("Unterminated ' in include", context)

      # remove quotes
      path = path[1:-1]

      # Probably a better way to do that...
      if not os.path.splitext(path)[1]:
        mangled = True
        path = os.path.join(path, 'index.lua')

    if len(args) > terminated and args[terminated] == 'as':
      terminated += 1

      if len(args) > terminated:
        aka = args[terminated]
        terminated += 1
      else:
        raise InvalidSyntaxProcessorException('Invalid import, missing name after \'as\'', context)

    # process remainder
    for i in range(terminated, len(args)):
      items = args[i].split(',')

      if items[0].startswith('+'):
        items[0] = items[0][1:]
        included = included.union(set(items))
      elif items[0].startswith('-'):
        items[0] = items[0][1:]
        excluded = excluded.union(set(items))
      else:
        raise_concern(f"Unrecognized group modifier prefix '{items[0][0]}', ignoring", context)

    if len(included.intersection(excluded)):
      raise_concern('Group names duplicated between include and exclude in include', context)

    # Handle relative paths
    if path.startswith('.'):
      # dirname is safe to do here since we added index.lua above for folders
      parent = os.path.dirname(context.path)
      path = os.path.join(parent, path)

    elif path and not path.startswith('/'):
      path = resolve_package(context, path, mangled)

    if not path:
      raise InvalidSyntaxProcessorException('Invalid include path; Did you forget quotes?', context)

    fullpath = os.path.abspath(path)

    if not fullpath in context.files:
      try:
        with open(path, 'r') as file:
          # Succeed in opening, *then* add to the list of files
          context.files.add(fullpath)

          return process_lines(ProcessorContext(
            path,
            file.read().split('\n'),
            aka=aka,
            included=included,
            excluded=excluded,
            context=context
          ))
      except FileNotFoundError:
        raise ModuleNotFoundProcessorException(f"Unable to locate included file '{path}'", context)
    else:
      # Future improvements could maybe seek to like remove other copies and replace them with this one if it would be a superset
      # of the original include
      raise_concern(f"File '{path}' included more than once; Did you include a file also included by your dependencies?", context)

  return None

def process_group_start(context, args):
  context.push_group(args[0])
  return None

def process_group_end(context, args):
  context.pop_group(args[0])
  return None

def process_exports(context, args):
  context.exports = set(args[0].split(','))
  return None

def process_import(context, args):
  if context.enabled:
    raise_concern('Import not supported. Use include.', context)
  return None

def process_unknown(context, args):
  if len(args) and args[0].startswith('#'):
    raise_concern(f"Unrecognized magic comment '{command}', skipping", context)
    return None
  return context.line

# https://www.lua.org/pil/2.4.html
supported_escapes = {
  'f': '\f',
  # New lines and carriage returns probably aren't supported in paths but...
  'n': '\n',
  'r': '\r',
  't': '\t',
  '\\': '\\',
  '"': '"',
  "'": "'",
  '[': '[',
  ']': ']'
}

# Only handles simple escapes, presumed ASCII so no \uXXXX
# Yes, there's built-in stuff to handle some forms of escape-encoded strings,
# but they all seem to have drawbacks:
# https://stackoverflow.com/questions/1885181/how-to-un-escape-a-backslash-escaped-string
def unescape(name, quote, context):
  unescaped_name = ''
  escaped = False

  for i in range(len(name)):
    char = name[i]

    if char == quote and not escaped:
      raise ProcessorException('Unsupported dynamic module name in require', context)

    if char == '\\' and not escaped:
      escaped = True
    elif escaped:
      if char in supported_escapes:
        unescaped_name += supported_escapes[char]
        escaped = False
      else:
        raise InvalidStringEscapeProcessorException('Invalid string escape in module name in require', context)
    else:
      unescaped_name += char

  return unescaped_name

# TODO This and all the other stuff needs to be replaced with proper AST stuff.
def parse_rvalue(remainder, context):
  rvalue = []
  part = ''
  quote = None
  escaped = False
  dot = False
  package_path_index = 0

  for i in range(len(remainder)):
    char = remainder[i]

    if escaped:
      if char in supported_escapes:
        part += char
        escaped = False
      else:
        raise InvalidStringEscapeProcessorException(f'Invalid string escape in {PACKAGE_PATH_VAR} assignment', context)
    elif char == quote:
      rvalue.append(part)
      part = ''
      quote = None
    elif char == '\\':
      escaped = True
    elif quote:
      part += char
    elif char == '"' or char == "'":
      quote = char
    elif char == '.' and (package_path_index == 0 or package_path_index == len(PACKAGE_PATH_VAR)):
      if dot:
        # Don't append '..' as we're doing the work to determine it's valid and don't want it in the output
        dot = False
      else:
        # Should just be package.path at this point
        if len(part):
          if part == PACKAGE_PATH_VAR:
            rvalue.append(part)
            part = ''
            package_path_index = 0
          else:
            raise DynamicAssignmentProcessorException(f'Unsupported dynamic {PACKAGE_PATH_VAR} assignment', context)
        dot = True
    elif char == ';':
      # This would be a syntax error trap below in all cases... I think?
      if len(part):
        rvalue.append(part)
        part = ''
        break
    elif package_path_index < len(PACKAGE_PATH_VAR) and char == PACKAGE_PATH_VAR[package_path_index]:
      part += PACKAGE_PATH_VAR[package_path_index]
      package_path_index += 1
    elif char == '=':
      raise InvalidSyntaxProcessorException(f'Invalid syntax in {PACKAGE_PATH_VAR} assignment', context)
    elif not re.match(r'\s', char):
      # Found a variable that's not package.path
      if i > 0 and not re.match(r'\s', remainder[i - 1]):
        raise DynamicAssignmentProcessorException(f'Unsupported dynamic {PACKAGE_PATH_VAR} assignment', context)

      # Lua lines can have multiple statements separated by a space instead of a semicolon.
      # Such fun.
      i = i - 1
      break

  if rvalue[:-1] == '..' or dot or quote or escaped:
    raise InvalidSyntaxProcessorException(f'Invalid syntax in {PACKAGE_PATH_VAR} assignment', context)

  # Line ends with package.path
  if part == PACKAGE_PATH_VAR:
    rvalue.append(PACKAGE_PATH_VAR)

  return (rvalue, i + 1)

# args could just be 'match' but this better aligns with the params passed to
# other processing functions line could also be passed to them, but they don't
# need it (yet)
def process_require(context, args, line):
  match = args[0]
  value = match.group(1)
  name = None
  quote = None

  if len(value):
    quote = value[0]
    endquote = value[-1]

    # We could scan this to check for the concatenation operator (..) but that's
    # also a valid parent path indicator so...
    if quote != endquote or not (quote == "'" or quote == '"'):
      raise DynamicModuleProcessorException('Unsupported dynamic module name in require', context)

    # strip quotes
    name = unescape(value[1:-1], quote, context)

  if not name:
    raise ModuleNotFoundProcessorException('Missing module name in require', context)

  require = context.requires.add_require(name)
  return line[0:match.start(0)] + require + '()' + line[match.end(0):]

def process_package_path(context, args, line):
  start = args[0].find('=') + 1
  parts, last_index = parse_rvalue(args[0][start:], context)
  path = context.requires.get_path()

  for i in range(len(parts)):
    part = parts[i].strip()

    if part == PACKAGE_PATH_VAR:
      if i > 0:
        context.requires.append_path(path)
    elif i == 0:
      context.requires.set_path(part)
    else:
      context.requires.append_path(part)

  
  # Strip lines that are empty after processing the package.path
  line = (line[0:args[1]] + line[args[1] + start + last_index:]).strip()

  return line if len(line) > 0 else None

def process_lines(context):
  output = []

  for i in range(len(context.lines)):
    context.line = context.lines[i]
    context.number = i
    processed = context.line

    if context.line.strip().startswith('--'):
      parts = context.line.strip().split()
      directive = parts[1] if len(parts) > 1 else ''
      args = parts[2:] if len(parts) > 2 else []

      # This could be refactored into a map -> processor since they currently
      # take all the same arguments and with the context, there's no real reason
      # for any to take different arguemnts except for convenience
      match directive:
        case MagicComment.GROUP_START:
          processed = process_group_start(context, args)

        case MagicComment.GROUP_END:
          processed = process_group_end(context, args)

        case MagicComment.IMPORT:
          processed = process_import(context, args)

        case MagicComment.INCLUDE:
          processed = process_include(context, args)

        case MagicComment.EXPORTS:
          processed = process_exports(context, args)

        case _:
          processed = process_unknown(context, args)

    elif context.line.find('require') != -1:
      match = re.search(r'require\(([^)]*)\)', context.line)

      if match:
        processed = process_require(context, [match], context.line)

    elif (index := context.line.find(PACKAGE_PATH_VAR)) != -1 and index < context.line.find('='):
      processed = process_package_path(context, [context.line[context.line.find(PACKAGE_PATH_VAR):], index], context.line)

    # We want to preserve line breaks from the input, but not ones that would be
    # caused by our preprocessing.
    if not context.disabled and not processed is None:
      output.append(processed)

  output = '\n'.join([line for line in output if not line is None])

  if context.requires.exists():
    output = context.requires.write() + '\n' + output

  if len(context.exports) and context.aka:
    output = context.aka + '={__autogen=true}\n' + output

    # FIXME This really needs a proper parser to get context and determine if we should be replacing something...
    for export in context.exports:
      output = re.sub(
        r'(?:(?<=\n)' + export + r'(?=\W))|(?:(?<![\w.])' + export + r'(\.))|(?:(?<=function )' + export + r'(?=\W))',
        f'{context.aka}.{export}\\1', output
      )

  return output




import os
import pytest
import unittest

from syllogistic.processor import ProcessorContext, ProcessorException, ModuleNotFoundProcessorException, InvalidSyntaxProcessorException, process_lines


class TestInclude(unittest.TestCase):
    def test_include(self):
      code = "-- #include './sample.lua'"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
      result = process_lines(context)

      assert result == """function sample()
  trace('sample')
end"""

    def test_include_absolute(self):
      path = os.path.abspath('tests/input/sample.lua')
      path = os.path.splitdrive(path)

      if (path[0] != ''):
        path = ('', path[1].replace('\\', '/'))

      code = f"-- #include '{path[1]}'"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
      result = process_lines(context)

      assert result == """function sample()
  trace('sample')
end"""

    def test_include_index(self):
      code = "-- #include './lib/test'"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
      result = process_lines(context)

      assert result == """-- lib/test/index.lua
-- aa.lua
aa={}

function aa.doA()
end
-- bb.lua
bb={
  doB=function()
  end
}
-- cc.lua
cc={
  doC=function ()
  end,
  doCGroupAOpt=function ()
  end,
}
"""

    def test_include_as(self):
      code = "-- #include './lib'"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
      result = process_lines(context)

      assert result == """-- lib/index.lua
lib={__autogen=true}
-- lib/test/index.lua
-- lib.aa.lua
lib.aa={}

function lib.aa.doA()
end
-- lib.bb.lua
lib.bb={
  doB=function()
  end
}
-- lib.cc.lua
lib.cc={
  doC=function ()
  end,
  doCGroupAOpt=function ()
  end,
}

"""

    def test_include_groups_remove(self):
      code = "-- #include './groups.lua' -C"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
      result = process_lines(context)

      assert result == """-- groups.lua
test={
  test=function ()
  end,
  testGroupA=function ()
  end,
}

"""

    def test_include_groups_remove2(self):
      code = "-- #include './groups.lua' -A"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
      result = process_lines(context)

      assert result == """-- groups.lua
test={
  test=function ()
  end,
  testGroupC=function ()
  end,
}

"""

    def test_include_groups_add(self):
      code = "-- #include './groups.lua' +B"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
      result = process_lines(context)

      assert result == """-- groups.lua
test={
  test=function ()
  end,
  testGroupA=function ()
  end,
  testGroupB=function ()
  end,
  testGroupC=function ()
  end,
}

function test.testGroupBStandalone()
end
"""

      code = "-- #include './groups.lua' +C"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
      result = process_lines(context)

      assert result == """-- groups.lua
test={
  test=function ()
  end,
  testGroupA=function ()
  end,
  testGroupC=function ()
  end,
}


function test.testGroupCStandalone()
end"""

    def test_include_groups_add2(self):
      code = "-- #include './groups.lua' +B,C"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
      result = process_lines(context)

      assert result == """-- groups.lua
test={
  test=function ()
  end,
  testGroupA=function ()
  end,
  testGroupB=function ()
  end,
  testGroupC=function ()
  end,
}

function test.testGroupBStandalone()
end

function test.testGroupCStandalone()
end"""

    def test_include_groups_add_remove_nest(self):
      code = "-- #include './groups.lua' -A +B"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
      result = process_lines(context)

      assert result == """-- groups.lua
test={
  test=function ()
  end,
  testGroupC=function ()
  end,
}

function test.testGroupBStandalone()
end
"""

      code = "-- #include './groups.lua' +B -A"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
      result = process_lines(context)

      assert result == """-- groups.lua
test={
  test=function ()
  end,
  testGroupC=function ()
  end,
}

function test.testGroupBStandalone()
end
"""

    def test_include_package(self):
      code = "-- #include 'test'"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
      result = process_lines(context)

      assert result == "debug={value='success'}"

    def test_include_package_part(self):
      code = "-- #include 'test/debug.lua'"
      context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
      result = process_lines(context)

      assert result == "debug={value='success'}"

    def test_include_no_win_absolute(self):
      with pytest.raises(ProcessorException):
        code = "-- #include 'C:\\tests\\input\\sample.lua'"
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
        result = process_lines(context)

    def test_include_malformed_include(self):
      with pytest.raises(InvalidSyntaxProcessorException):
        code = "-- #include ./sample.lua"
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
        result = process_lines(context)

    def test_include_disallowed_prefix(self):
      with pytest.raises(ProcessorException):
        code = "-- #include '~sample.lua'"
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
        result = process_lines(context)

    def test_include_malformed_as(self):
      with pytest.raises(InvalidSyntaxProcessorException):
        code = "-- #include './sample.lua' as"
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
        result = process_lines(context)

    def test_include_unknown_package(self):
      with pytest.raises(ModuleNotFoundProcessorException):
        code = "-- #include 'thisisnotavalidpackagenameandlikelywontexist'"
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
        result = process_lines(context)

    def test_include_malformed_groups(self):
      with pytest.raises(InvalidSyntaxProcessorException):
        code = "-- #include ./groups-bad.lua"
        context = ProcessorContext('tests/input/fake.lua', code.split('\n'), strict=True);
        result = process_lines(context)



import math
import os
import time

from time import time, strftime
from datetime import datetime
from pakettic import ticfile
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from syllogistic.processor import ProcessorContext, ProcessorException, process_lines

class BuildWatcher(FileSystemEventHandler):
  files = set()
  dirs = set()
  watches = dict()

  def __init__(self, version, file, outfile, strict, default_require_resolution, verbose):
    self.version = version
    self.file = os.path.abspath(file)
    self.outfile = outfile
    self.strict = strict
    self.default_require_resolution = default_require_resolution
    self.verbose = verbose
    self.observer = Observer()
    self.last_src = None
    self.last_time = time()

  def _build(self):
    dirs = set()
    context = self.build(True)

    if context:
      self.files = context.files
      self.files.add(self.file)

      for file in self.files:
        # Assumed to already be absolute paths
        dirs.add(os.path.dirname(file))

      for dir in self.dirs.difference(dirs):
        if dir in self.watches:
          self.observer.unschedule(self.watches[dir])
          del self.watches[dir]

      for dir in dirs.difference(self.dirs):
        self.watches[dir] = self.observer.schedule(self, dir)

      self.dirs = dirs

  def build(self, watched=False):
    context = None

    try:
      start = time()
      beginmsg = f'\r[{strftime("%H:%M")}] File updated, building...'
      print(beginmsg, end='')

      cart = ticfile.join_code(ticfile.read(self.file))

      for chunk in [chunk for chunk in cart if chunk[1] == ticfile.ChunkID.CODE]:
        code = cart[chunk].decode('ascii')
        context = ProcessorContext(self.file, code.split('\n'), strict=self.strict, watched=True, verbose=self.verbose);
        result = process_lines(context)
        result = f'-- builder: SyllogisTIC v{self.version}\n-- built: {datetime.now()}Z\n' + result
        cart[chunk] = result.encode('ascii')
        ticfile.write(cart, self.outfile)

      end = (time() - start) * 1000
      end = math.floor(end) / 1000 if end > 1000 else str(math.floor(end)) + 'm'
      print(f'\r[{strftime("%H:%M")}] Built in {end}s'.ljust(len(beginmsg), ' '), end='')

    except ProcessorException as pe:
      print(f'\nError: {pe.args[0]}; in {pe.file} on line {pe.line}')
      print(f'[{strftime("%H:%M")}] Build failed with errors')

    return context

  def start(self):
    print('')
    self._build()
    self.observer.start()

    try:
      while self.observer.is_alive():
        self.observer.join(1)
    except KeyboardInterrupt:
      print('\nAsked to quit. Goodbye...')
    finally:
      self.stop()

  def stop(self):
    self.observer.stop()
    try:
      self.observer.join()
    except:
      pass

  def on_deleted(self, event):
    if event.src_path == self.file:
      print('\nInput file deleted. Ending watch.')
      self.stop()

  def on_modified(self, event):
    now = time()

    # 10ms is probably adequate for debouncing if we get called more than once for the same modification
    # The author's Windows 10 Pro is seeing ~5-6ms between the first and second of the doubled event
    if event.src_path in self.files and (event.src_path != self.last_src or (now - self.last_time) * 1000 > 10):
      self.last_src = event.src_path
      self.last_time = now
      self.build()

  def on_moved(self, event):
    if event.src_path == self.file:
      # Technically we could update self.file and keep on going,
      # but things could get weird depending on where it's moved to
      print('\nInput file moved. Ending watch.')
      self.stop()
